"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
var tslib_1 = require("tslib");
var errors_1 = tslib_1.__importDefault(require("../errors"));
var utils = tslib_1.__importStar(require("../utils"));
var TableSchemaModel = /** @class */ (function () {
    function TableSchemaModel(db, name, definition) {
        var _this = this;
        this._relations = [];
        this.db = utils.ensureParamObject("db", db);
        this.name = utils.ensureParamString("name", name);
        utils.ensureParamObject("definition", definition);
        this.fields = Object.keys(definition)
            .map(function (fieldName) {
            return db.factory.newFieldSchema(_this, fieldName, definition[fieldName]);
        });
        this._normalizer = db.factory.newSchemaNormalizer(this);
        this._primaryKeyFields = this.fields.filter(function (f) { return f.isPrimaryKey; });
        this._foreignKeyFields = this.fields.filter(function (f) { return f.isForeignKey; });
        this._stampFields = this.fields.filter(function (f) { return f.isStamp; });
    }
    Object.defineProperty(TableSchemaModel.prototype, "relations", {
        get: function () { return this._relations; },
        enumerable: true,
        configurable: true
    });
    Object.defineProperty(TableSchemaModel.prototype, "primaryKeys", {
        get: function () { return this._primaryKeyFields; },
        enumerable: true,
        configurable: true
    });
    TableSchemaModel.prototype.connect = function (tableMap) {
        var _this = this;
        utils.ensureParamObject("tableMap", tableMap);
        Object.keys(tableMap).forEach(function (schema) {
            return _this._relations = _this._relations.concat(tableMap[schema].fields.filter(function (f) { return f.references === _this.name; }));
        });
        this._foreignKeyFields.forEach(function (fk) { return fk.connect(tableMap); });
    };
    TableSchemaModel.prototype.inferRelations = function (data, rel, ownerId) {
        if (!rel.isForeignKey)
            return data;
        var otherFks = rel.table.fields.filter(function (f) { return f.isForeignKey && f !== rel; });
        return utils.ensureArray(data).map(function (obj) {
            var _a, _b;
            if (typeof obj === "number" || typeof obj === "string")
                if (otherFks.length === 1)
                    obj = (_a = {}, _a[otherFks[0].name] = obj, _a);
                else
                    obj = { id: obj }; // TODO: this might be quite wrong..
            return tslib_1.__assign({}, obj, (_b = {}, _b[rel.name] = ownerId, _b));
        });
    };
    TableSchemaModel.prototype.normalize = function (data, context) {
        return this._normalizer.normalize(data, context);
    };
    TableSchemaModel.prototype.injectKeys = function (data, record) {
        if (!data || typeof data !== "object")
            return data;
        // inject primary or foreign keys
        var keys = this._primaryKeyFields;
        if (!keys.length)
            keys = this._foreignKeyFields;
        keys.forEach(function (key) {
            if (data[key.name] === undefined)
                data[key.name] = key.getRecordValue(record);
        });
    };
    TableSchemaModel.prototype.ensurePrimaryKey = function (record) {
        var pk = this.getPrimaryKey(record);
        if (!pk)
            throw new Error(errors_1.default.pkNotFound(this.name));
        return pk;
    };
    TableSchemaModel.prototype.getPrimaryKey = function (record) {
        var lookup = (this._primaryKeyFields.length ? this._primaryKeyFields : this._foreignKeyFields);
        var combinedPk = lookup.reduce(function (p, n) {
            var k = n.getValue(record);
            return p && k ? (p + "_" + k) : k;
        }, null);
        if (utils.isValidID(combinedPk))
            return utils.asID(combinedPk);
        else
            return undefined;
    };
    TableSchemaModel.prototype.getForeignKeys = function (record) {
        var _this = this;
        return this._foreignKeyFields.map(function (fk) {
            if (!fk.references || !fk.refTable)
                throw new Error(errors_1.default.fkInvalid(_this.name, fk.name));
            return {
                name: fk.name,
                notNull: fk.notNull,
                references: fk.references,
                refTable: fk.refTable,
                unique: fk.unique === true,
                value: record[fk.name]
            };
        });
    };
    TableSchemaModel.prototype.isModified = function (x, y) {
        if (this._stampFields.length > 0)
            return this._stampFields.reduce(function (p, n) {
                return p + (n.getValue(x) === n.getValue(y) ? 1 : 0);
            }, 0) !== this._stampFields.length;
        else {
            var comparer = this.db.getRecordComparer(this.name);
            if (comparer)
                return !comparer(x, y, this);
            return x === y;
        }
    };
    TableSchemaModel.prototype.mergeRecord = function (oldRecord, newRecord) {
        var merger = this.db.getRecordMerger(this.name);
        if (merger)
            return merger(oldRecord, newRecord, this);
        return tslib_1.__assign({}, oldRecord, newRecord);
    };
    return TableSchemaModel;
}());
exports.default = TableSchemaModel;
