declare module "types" {
    export type FieldType = "ATTR" | "MODIFIED" | "PK";
    export type ValueType<R> = R extends {
        value: infer ValueType;
    } ? ValueType : never;
    export type Values<R> = ValueType<R> | ValueType<R>[];
    export type PartialValue<R> = Partial<ValueType<R>>;
    export type PartialValues<R> = PartialValue<R> | (PartialValue<R>[]);
    export interface RecordEntity {
        id: string;
    }
    export type Reducer = (session: any, action: any, arg?: any) => void;
    export type PkGenerator = (record: any, schema: TableSchema) => string | null | undefined;
    export type RecordNormalizer = (record: any, context: NormalizeContext) => any;
    export type RecordComparer = (a: any, b: any, schema: TableSchema) => boolean;
    export type RecordMerger = (a: any, b: any, schema: TableSchema) => any;
    export interface MapOf<T> {
        [key: string]: T;
    }
    export interface ForeignKey {
        name: string;
        value: string;
        references: string;
        refTable: TableSchema;
        unique: boolean;
        notNull: boolean;
    }
    export interface TableSchema {
        db: DatabaseSchema;
        name: string;
        fields: FieldSchema[];
        relations: FieldSchema[];
        primaryKeys: FieldSchema[];
        connect(schemas: MapOf<TableSchema>): void;
        ensurePrimaryKey(record: any): string;
        getPrimaryKey(record: any): string | undefined;
        getForeignKeys(record: any): ForeignKey[];
        isModified(recordA: any, recordB: any): boolean;
        inferRelations(data: any, rel: FieldSchema, ownerId: string): any[];
        injectKeys(data: any, record: TableRecord): void;
        normalize(data: any, context: NormalizeContext): string[];
        mergeRecord(oldRecord: any, newRecord: any): any;
    }
    export interface FieldSchema {
        table: TableSchema;
        refTable?: TableSchema;
        name: string;
        propName: string;
        references?: string;
        relationName?: string;
        cascade?: boolean;
        unique?: boolean;
        isPrimaryKey: boolean;
        isForeignKey: boolean;
        isStamp: boolean;
        notNull: boolean;
        connect(schemas: MapOf<TableSchema>): void;
        getValue(data: any, record?: any): any;
        getRecordValue(record: any): any;
    }
    export interface Table<R extends TableRecord = TableRecord> {
        session: Session;
        schema: TableSchema;
        state: TableState;
        dirty: boolean;
        get(id: string | number): R;
        getValue(id: string | number): ValueType<R> | undefined;
        getIndex(schemaName: string, fkId: string): string[];
        /**
         * Gets all records in table as wrapped models.
         *
         * @returns {R[]}
         * @memberof Table
         */
        all(): R[];
        /**
         * Gets all values in table.
         *
         * @returns {ValueType<R>[]}
         * @memberof Table
         */
        values(): ValueType<R>[];
        /**
         * Checks whether a record exists in table.
         *
         * @param {(string | number)} id
         * @returns {boolean}
         * @memberof Table
         */
        exists(id: string | number): boolean;
        /**
         * Inserts single or multiple records.
         *
         * @param {Values<R>} data
         * @param {*} [argument] Optional argument passed to normalization.
         * @returns {string[]} Primary keys of inserted records.
         * @memberof Table
         */
        insert(data: Values<R>, argument?: any): string[];
        /**
         * Updates single or multiple records.
         *
         * @param {PartialValues<R>} data
         * @param {*} [argument] Optional argument passed to normalization.
         * @returns {string[]} Primary keys of updated records.
         * @memberof Table
         */
        update(data: PartialValues<R>, argument?: any): string[];
        /**
         * Upserts single or multiple records.
         *
         * @param {PartialValues<R>} data
         * @param {*} [argument] Optional argument passed to normalization.
         * @returns {string[]} Primary keys of upserted records.
         * @memberof Table
         */
        upsert(data: PartialValues<R>, argument?: any): string[];
        /**
         * Deletes single or multiple records by id or object.
         *
         * @param {(string | number | PartialValue<R> | (string | number | PartialValue<R>)[])} data
         * @returns {number} Number of successfully deleted records.
         * @memberof Table
         */
        delete(data: string | number | PartialValue<R> | (string | number | PartialValue<R>)[]): number;
        /**
         * Deletes all records in table.
         *
         * @memberof Table
         */
        deleteAll(): void;
        /**
         * Upserts the table state from another normalized state.
         *
         * @param {TableState<ValueType<R>>} table
         * @memberof Table
         */
        upsertNormalized(table: TableState<ValueType<R>>): void;
        /**
         * Normalizes given data.
         *
         * @param {PartialValues<R>} data
         * @param {NormalizeOptions} [options]
         * @returns {NormalizeContext}
         * @memberof Table
         */
        normalize(data: PartialValues<R>, options?: NormalizeOptions): NormalizeContext;
    }
    /**
     * Represents a wrapped record belonging to a table.
     *
     * @export
     * @interface TableRecord
     * @template T
     */
    export interface TableRecord<T = any> {
        /**
         * Gets the record primary key.
         *
         * @type {string}
         * @memberof TableRecord
         */
        id: string;
        /**
         * Gets the table model which this records belongs.
         *
         * @type {Table}
         * @memberof TableRecord
         */
        table: Table;
        /**
         * Gets or sets the record data value.
         *
         * @type {T}
         * @memberof TableRecord
         */
        value: T;
        /**
         * Update record with new data.
         *
         * @param {Partial<T>} data - Object with partial or full data.
         * @returns {this} Self
         * @memberof TableRecord
         */
        update(data: Partial<T>): this;
        /**
         * Deletes the record from table.
         *
         * @memberof TableRecord
         */
        delete(): void;
    }
    export interface TableRecordField {
        name: string;
        value: any;
    }
    export interface TableRecordSet<R extends TableRecord = TableRecord> {
        ids: string[];
        length: number;
        all(): R[];
        values(): ValueType<R>[];
        add(data: Values<R>): string[];
        remove(data: PartialValue<R>): number;
        update(data: PartialValues<R>): string[];
        delete(): void;
    }
    export interface SchemaNormalizer {
        normalize: (data: any, context: NormalizeContext) => string[];
    }
    export interface ModelFactory {
        newSchemaNormalizer: (schema: TableSchema) => SchemaNormalizer;
        newSession: (db: DatabaseSchema, state: DatabaseState, options?: SessionOptions) => Session;
        newTableSchema: (db: DatabaseSchema, name: string, schema: TableDefinition) => TableSchema;
        newFieldSchema: (table: TableSchema, name: string, schema: FieldDefinition) => FieldSchema;
        newTableModel: (session: Session, schema: TableSchema, state: TableState) => Table;
        newRecordSetModel: (table: Table, schema: FieldSchema, owner: RecordEntity) => TableRecordSet;
        newRecordFieldModel: (schema: FieldSchema, record: TableRecord) => TableRecordField;
    }
    export interface RecordFactory {
        newRecordModel: (id: string, table: Table) => TableRecord;
    }
    /**
     * Defines a database schema
     */
    export interface Schema {
        [key: string]: TableDefinition;
    }
    /**
     * Defines a table schema
     */
    export interface TableDefinition {
        [key: string]: FieldDefinition;
    }
    export interface FieldDefinition {
        /**
         * The field type (Deprecated)
         *
         * @type {FieldType}
         * @memberof FieldDefinition
         * @deprecated
         */
        type?: FieldType;
        pk?: boolean;
        stamp?: boolean;
        /**
         * The name this field references´ in the model. Defaults to schema key.
         *
         * @type {string}
         * @memberof FieldDefinition
         */
        fieldName?: string;
        propName?: string;
        references?: string;
        relationName?: string;
        cascade?: boolean;
        unique?: boolean;
        notNull?: boolean;
        value?: (record: any, context: ComputeContext<any>) => any;
    }
    export interface ComputeContext<T> {
        schema: FieldSchema;
        record: TableRecord<T>;
    }
    export interface NormalizeContext {
        table?: Table;
        schema: TableSchema;
        db: DatabaseSchema;
        output: MapOf<TableState>;
        emits: {
            [key: string]: any[];
        };
        normalizePKs: boolean;
        emit(tableName: string, record: any): void;
    }
    export interface DatabaseSchema {
        tables: TableSchema[];
        options: DatabaseOptions;
        factory: ModelFactory & RecordFactory;
        getRecordNormalizer: (schemaName: string) => RecordNormalizer | undefined;
        getPkGenerator: (schemaName: string) => PkGenerator | undefined;
        getRecordComparer: (schemaName: string) => RecordComparer | undefined;
        getRecordMerger: (schemaName: string) => RecordMerger | undefined;
    }
    export interface DatabaseOptions {
        onNormalize?: MapOf<RecordNormalizer> | RecordNormalizer;
        onGeneratePK?: MapOf<PkGenerator> | PkGenerator;
        onRecordCompare?: MapOf<RecordComparer> | RecordComparer;
        onRecordMerge?: MapOf<RecordMerger> | RecordMerger;
        cascadeAsDefault?: boolean;
        strict?: boolean;
    }
    export interface DatabaseCreateOptions extends DatabaseOptions {
        factory?: Partial<ModelFactory>;
        recordModelClass?: RecordClass;
    }
    export interface SessionOptions {
        readOnly?: boolean;
        tableSchemas?: TableSchema[];
    }
    export interface NormalizeOptions {
        normalizePKs?: boolean;
        argument?: any;
    }
    export interface NormalizeContextOptions extends NormalizeOptions {
        table?: Table;
    }
    export type DatabaseState = MapOf<TableState>;
    export type TableMap = MapOf<Table>;
    export interface TableState<T = any> {
        name?: string;
        byId: MapOf<T>;
        ids: string[];
        indexes: MapOf<TableIndex>;
    }
    export interface TableIndex {
        unique: boolean;
        values: MapOf<string[]>;
    }
    export interface Session<T extends TableMap = TableMap> {
        db: DatabaseSchema;
        state: DatabaseState;
        tables: T;
        readOnly: boolean;
        upsert(ctx: NormalizeContext): void;
        commit(): DatabaseState;
    }
    export interface RecordClass {
        new (id: string, table: Table): TableRecord;
    }
    export interface ExtendedRecord extends TableRecord {
        _fields: {
            [key: string]: any;
        };
    }
}
declare module "constants" {
    import { TableState } from "types";
    export const TYPE_PK = "PK";
    export const TYPE_ATTR = "ATTR";
    export const TYPE_MODIFIED = "MODIFIED";
    export const ALL = "*";
    export const RESERVED_PROPERTIES: string[];
    export const initialState: <T = any>(name?: string | undefined) => TableState<T>;
}
declare module "errors" {
    const _default: {
        argument: (name: string, type: string) => string;
        argumentShape: (name: string, props: string[]) => string;
        fkInvalid: (table: string, key: string) => string;
        fkInvalidReference: (table: string, key: string, reference: string) => string;
        fkReferenceNotInSession: (key: string, references?: string | undefined) => string;
        fkUndefined: (table: string, key: string) => string;
        fkViolation: (table: string, key: string) => string;
        invalidId: () => string;
        normalizeInvalidData: () => string;
        normalizeInvalidFk: (table: string, key: string, refTable: string) => string;
        normalizePk: (name: string) => string;
        normalizeFkViolation: (table: string, key: string) => string;
        pkNotFound: (table: string) => string;
        recordNotFound: (table: string, id: string | number) => string;
        recordUpdateViolation: (table: string, id: string) => string;
        reservedProperty: (name: string, prop: string) => string;
        sessionReadonly: () => string;
        stateTableUndefined: () => string;
        tableInvalidState: (table: string) => string;
        tableNotInSession: (table: string) => string;
        uniqueConstraintViolation: (id: string) => string;
        unknownTableState: () => string;
    };
    export default _default;
}
declare module "utils" {
    export const toArray: (obj: any) => any[];
    export const ensureArray: (obj: any) => any[];
    export const isObject: (value: any) => boolean;
    export const isPlainObject: (value: any) => boolean;
    export const ensureParam: <T = any>(name: string, value: T) => T;
    export const ensureParamString: (name: string, value: any) => string;
    export const ensureParamObject: (name: string, value: any, ...props: string[]) => any;
    export const ensureParamFunction: (name: string, value: any) => Function;
    export function optionalParamString(name: string, val: any, fallback: string): string;
    export function optionalParamString(name: string, val: any): string | undefined;
    export const ensureID: (id: any) => string;
    export const isValidID: (id: any) => boolean;
    export const asID: (id: any) => string;
    export const toObject: <T>(a: T[], key: (a: T) => string) => Record<string, T>;
    export const mergeIds: (source: string[], second: string[], unique: boolean) => string[];
    export const isEqual: (a: any, b: any) => boolean;
}
declare module "ModelFactory" {
    import { ModelFactory, RecordClass, RecordFactory, Table, TableRecord, TableSchema } from "types";
    export const createRecordModelClass: (Base: RecordClass) => {
        new (id: string, table: Table<TableRecord<any>>): {
            _fields: {
                [key: string]: any;
            };
            id: string;
            table: Table<TableRecord<any>>;
            value: any;
            update(data: Partial<any>): any;
            delete(): void;
        };
    };
    export class RecordModelFactory {
        private _recordClass;
        private _recordBaseType;
        private _factory;
        constructor(recordBaseType: RecordClass, factory: ModelFactory);
        newRecordModel(id: string, table: Table): TableRecord<any>;
        protected createRecordModel(schema: TableSchema): RecordClass;
        private _newRecordField;
        private _newRecordSet;
        private _newRecordRelation;
        private _defineProperty;
    }
    const createModelFactory: (factory: ModelFactory, RecordModel: RecordClass) => ModelFactory & RecordFactory;
    export default createModelFactory;
}
declare module "Normalizer" {
    import { DatabaseSchema, FieldSchema, ForeignKey, NormalizeContext, TableSchema, TableState } from "types";
    export default class SchemaNormalizer {
        schema: TableSchema;
        db: DatabaseSchema;
        constructor(schema: TableSchema);
        normalize(data: any, context: NormalizeContext): string[];
        protected _normalizePrimaryKey(record: any): string | null | undefined;
        protected _normalizeRelations(rel: FieldSchema, record: any, pk: string, context: NormalizeContext): void;
        protected _normalizeForeignKey(fk: ForeignKey, record: any, context: NormalizeContext): void;
        protected _indexForeignKey(fk: ForeignKey, tbl: TableState, pk: string): void;
    }
}
declare module "models/Database" {
    import { DatabaseOptions, DatabaseSchema, DatabaseState, MapOf, ModelFactory, RecordFactory, Reducer, Schema, Session, SessionOptions, TableMap, TableSchema } from "types";
    export default class Database<T extends TableMap> implements DatabaseSchema {
        schema: Schema;
        tables: TableSchema[];
        options: DatabaseOptions;
        factory: ModelFactory & RecordFactory;
        tableMap: MapOf<TableSchema>;
        constructor(schema: Schema, factory: ModelFactory & RecordFactory, options?: DatabaseOptions);
        getRecordNormalizer: (schemaName: string) => import("types").RecordNormalizer | undefined;
        getPkGenerator: (schemaName: string) => import("types").PkGenerator | undefined;
        getRecordComparer: (schemaName: string) => import("types").RecordComparer | undefined;
        getRecordMerger: (schemaName: string) => import("types").RecordMerger | undefined;
        combineReducers(...reducers: Reducer[]): (state: any, action: any) => MapOf<import("types").TableState<any>>;
        reduce(state?: DatabaseState, action?: any, reducers?: Reducer | Reducer[], ...args: any[]): MapOf<import("types").TableState<any>>;
        createSession(state?: DatabaseState, options?: SessionOptions): Session<T>;
        getTableSchema(name: string): TableSchema;
        selectTables(state: any): T;
        selectTable(tableState: any, schemaName?: string): import("types").Table<import("types").TableRecord<any>>;
    }
}
declare module "models/DatabaseSession" {
    import { DatabaseSchema, DatabaseState, NormalizeContext, Session, SessionOptions, TableMap } from "types";
    export default class DatabaseSession implements Session {
        readonly db: DatabaseSchema;
        readonly tables: TableMap;
        state: DatabaseState;
        readOnly: boolean;
        constructor(db: DatabaseSchema, state: DatabaseState, options?: SessionOptions);
        getTable(name: string): import("types").Table<import("types").TableRecord<any>>;
        upsert(ctx: NormalizeContext): void;
        commit(): import("types").MapOf<import("types").TableState<any>>;
    }
}
declare module "models/FieldSchemaModel" {
    import { FieldDefinition, FieldSchema, FieldType, MapOf, TableSchema } from "types";
    export default class FieldSchemaModel implements FieldSchema {
        readonly table: TableSchema;
        readonly name: string;
        readonly propName: string;
        readonly type: FieldType;
        readonly references?: string;
        readonly relationName?: string;
        readonly cascade: boolean;
        readonly unique: boolean;
        readonly notNull: boolean;
        readonly isPrimaryKey: boolean;
        readonly isForeignKey: boolean;
        readonly isStamp: boolean;
        private _refTable?;
        private _valueFactory?;
        constructor(table: TableSchema, name: string, schema: FieldDefinition, cascadeAsDefault?: boolean);
        readonly refTable: TableSchema | undefined;
        connect(schemas: MapOf<TableSchema>): void;
        getValue(data: any, record?: any): any;
        getRecordValue(record: {
            value: any;
        }): any;
    }
}
declare module "models/RecordFieldModel" {
    import { FieldSchema, TableRecord, TableRecordField } from "types";
    export default class RecordFieldModel<T> implements TableRecordField {
        readonly record: TableRecord<T>;
        readonly schema: FieldSchema;
        readonly name: string;
        constructor(schema: FieldSchema, record: TableRecord<T>);
        readonly value: any;
    }
}
declare module "models/RecordModel" {
    import { Table, TableRecord } from "types";
    export default class RecordModel<T> implements TableRecord<T> {
        table: Table<RecordModel<T>>;
        id: string;
        constructor(id: string, table: Table);
        value: T;
        delete(): void;
        update(data: Partial<T>): this;
    }
}
declare module "models/RecordSetModel" {
    import { FieldSchema, PartialValues, Table, TableRecord, TableRecordSet, Values } from "types";
    export default class RecordSetModel<R extends TableRecord> implements TableRecordSet<R> {
        readonly table: Table<R>;
        readonly schema: FieldSchema;
        readonly owner: {
            id: string;
        };
        constructor(table: Table<R>, schema: FieldSchema, owner: {
            id: string;
        });
        readonly ids: string[];
        readonly length: number;
        all(): R[];
        values(): import("types").ValueType<R>[];
        add(data: Values<R>): string[];
        remove(data: PartialValues<R>): number;
        update(data: PartialValues<R>): string[];
        delete(): void;
        private _normalize;
    }
}
declare module "state" {
    import { ForeignKey, MapOf, TableIndex, TableState } from "types";
    export const merge: (original: TableState<any>, modified: Partial<TableState<any>>) => {
        byId: MapOf<any>;
        ids: string[];
        name?: string | undefined;
        indexes: MapOf<TableIndex>;
    };
    export const splice: (original: TableState<any>, idsToDelete: string[]) => {
        deleted: {
            id: string;
            record: any;
        }[];
        state: {
            byId: {
                [x: string]: any;
            };
            ids: string[];
            indexes: {
                [x: string]: TableIndex;
            };
            name?: string | undefined;
        };
    };
    export const updateIndexes: (name: string, original: TableState<any>, modified: TableState<any>) => void;
    export const cleanIndexes: (keys: ForeignKey[], id: string, indexes: MapOf<TableIndex>) => void;
    const _default_1: {
        cleanIndexes: (keys: ForeignKey[], id: string, indexes: MapOf<TableIndex>) => void;
        merge: (original: TableState<any>, modified: Partial<TableState<any>>) => {
            byId: MapOf<any>;
            ids: string[];
            name?: string | undefined;
            indexes: MapOf<TableIndex>;
        };
        splice: (original: TableState<any>, idsToDelete: string[]) => {
            deleted: {
                id: string;
                record: any;
            }[];
            state: {
                byId: {
                    [x: string]: any;
                };
                ids: string[];
                indexes: {
                    [x: string]: TableIndex;
                };
                name?: string | undefined;
            };
        };
        updateIndexes: (name: string, original: TableState<any>, modified: TableState<any>) => void;
    };
    export default _default_1;
}
declare module "models/NormalizeContext" {
    import { DatabaseSchema, MapOf, NormalizeContext, NormalizeContextOptions, Table, TableSchema, TableState } from "types";
    export default class DbNormalizeContext implements NormalizeContext {
        table?: Table;
        schema: TableSchema;
        db: DatabaseSchema;
        output: MapOf<TableState>;
        emits: {
            [key: string]: any[];
        };
        normalizePKs: boolean;
        argument?: any;
        constructor(schema: TableSchema, options: NormalizeContextOptions);
        emit(tableName: string, record: any): void;
    }
}
declare module "models/TableModel" {
    import { NormalizeOptions, PartialValue, PartialValues, Session, Table, TableRecord, TableSchema, TableState, Values, ValueType } from "types";
    import NormalizeContext from "models/NormalizeContext";
    export default class TableModel<R extends TableRecord> implements Table<R> {
        readonly session: Session;
        readonly schema: TableSchema;
        state: TableState<ValueType<R>>;
        dirty: boolean;
        constructor(session: Session, schema: TableSchema, state?: TableState);
        readonly length: number;
        all(): R[];
        values(): ValueType<R>[];
        exists(id: number | string): boolean;
        get(id: number | string): R;
        getValue(id: number | string): ValueType<R> | undefined;
        getIndex(name: string, fk: string): string[];
        insert(data: Values<R>, argument?: any): string[];
        update(data: PartialValues<R>, argument?: any): string[];
        upsert(data: PartialValues<R>, argument?: any): string[];
        delete(data: string | number | PartialValue<R> | (string | number | PartialValue<R>)[]): number;
        deleteAll(): void;
        upsertNormalized(norm: TableState<ValueType<R>>): void;
        insertNormalized(table: TableState<ValueType<R>>, updateIndexes?: boolean): void;
        updateNormalized(table: TableState<ValueType<R>>, updateIndexes?: boolean): void;
        normalize(data: PartialValues<R>, options?: NormalizeOptions): NormalizeContext;
        private _normalizedAction;
        private _cleanIndexes;
        private _updateIndexes;
        private _deleteCascade;
    }
}
declare module "models/TableSchemaModel" {
    import { DatabaseSchema, FieldSchema, MapOf, NormalizeContext, TableDefinition, TableRecord, TableSchema } from "types";
    export default class TableSchemaModel implements TableSchema {
        readonly db: DatabaseSchema;
        readonly name: string;
        readonly fields: FieldSchema[];
        private _normalizer;
        private _relations;
        private _primaryKeyFields;
        private _foreignKeyFields;
        private _stampFields;
        constructor(db: DatabaseSchema, name: string, definition: TableDefinition);
        readonly relations: FieldSchema[];
        readonly primaryKeys: FieldSchema[];
        connect(tableMap: MapOf<TableSchema>): void;
        inferRelations(data: any, rel: FieldSchema, ownerId: string): any[];
        normalize(data: any, context: NormalizeContext): string[];
        injectKeys(data: any, record: TableRecord): any;
        ensurePrimaryKey(record: any): string;
        getPrimaryKey(record: any): string | undefined;
        getForeignKeys(record: any): {
            name: string;
            notNull: boolean;
            references: string;
            refTable: TableSchema;
            unique: boolean;
            value: any;
        }[];
        isModified(x: any, y: any): boolean;
        mergeRecord(oldRecord: any, newRecord: any): any;
    }
}
declare module "models/index" {
    import Database from "models/Database";
    import DatabaseSession from "models/DatabaseSession";
    import FieldSchemaModel from "models/FieldSchemaModel";
    import RecordFieldModel from "models/RecordFieldModel";
    import RecordModel from "models/RecordModel";
    import RecordSetModel from "models/RecordSetModel";
    import TableModel from "models/TableModel";
    import TableSchemaModel from "models/TableSchemaModel";
    export { Database, DatabaseSession, FieldSchemaModel, RecordFieldModel, RecordModel, RecordSetModel, TableModel, TableSchemaModel };
}
declare module "index" {
    import { Database } from "models/index";
    import { DatabaseCreateOptions, ModelFactory, RecordClass, Schema } from "types";
    export const createFactory: (factory?: Partial<ModelFactory> | undefined, recordModelClass?: RecordClass | undefined) => ModelFactory & import("types").RecordFactory;
    export const createDatabase: <T extends import("types").MapOf<import("types").Table<import("types").TableRecord<any>>>>(schema: Schema, options?: DatabaseCreateOptions) => Database<T>;
    export * from "constants";
    export * from "models/index";
    export * from "ModelFactory";
    export * from "types";
    import * as utils from "utils";
    export { utils };
}
